/***********************************************************************

	video.c

//   ̃\[XR[h́A֓N玁@"NTSCrfIo͂̃IVXR[v"@̃vO
//  قƂǂ̂܂܎g킹ĒĂ܂B
//   MZ-80 emulater ɍ킹ĕύXĂ܂B
//
//  dqH̎uPIC24FpЉv
//   http://www.picfun.com/app24Fframe.html

	2013/10/30  S.Suwa http://www.suwa-koubou.jp
	
***********************************************************************/

#include "GenericTypeDefs.h"
#include "HardwareProfile.h"
#include "mz1200.h"

/*----------------------------------------------------------------------
	constant definition
----------------------------------------------------------------------*/

#define V_RES		200		// 25line*8dot vertical resolution
#define H_RES		320		// 40chr*8dot horizontal resolution

// video state
#define SV_PREEQ	0       // pre sync
#define SV_SYNC		1       // sync
#define SV_POSTEQ	2       // post sync
#define	SV_LINE		3       // video out

#define V_NTSC		262		// horizontal lines/frame 
#define V_SYNC		3 		// vertical sync
#define V_BLANK		(V_NTSC - V_RES - V_SYNC)
#define V_PREEQ		V_BLANK / 2	        // upper blanking(include presync)
#define V_POSTEQ	V_BLANK - V_PREEQ	// lower blanking(include postsync)
#define H_NTSC		2542   // 63.55usec Tcy 2542
                           // 40,000/15.734=2542.3-1=2541
                           // 40,000/2542=15.73564 (spec 15,734KHz)
                           // 25nsec*2542=63.55usec (spec 63.56usec)
#define H_SYNC		188	   // 4.7usec Tcy 188 
                           // 4.7/(1/40M)=188 
//#define H_BACK	188    // 4.7usec for dot clock 6.666MHz
#define H_BACK		340    // 4.7usec+nn for dot clock 8MHz


/*----------------------------------------------------------------------
	variable
----------------------------------------------------------------------*/

// Video memory address is from 0xD000 to 0xD7ff.
// This 2Kbyte memorys is used as ring buffer in MZ-80A. 
// Video screen hardware roll up, roll down, controled by monitor program.
static unsigned short VPtr;     // 0xD000-0xD7FF
static unsigned short Lptr;     // line begin address
volatile unsigned short VMem;   // Video memory begin address
static unsigned char *cgrom;

static volatile unsigned int  VCount, VState, DotCount, Line;
static unsigned char VS[4] = {SV_SYNC, SV_POSTEQ, SV_LINE, SV_PREEQ};
static unsigned char VC[4] = {V_SYNC, V_POSTEQ, V_RES, V_PREEQ};

// file select control
int bgnln, endln, bgndt, enddt, mnrvs;

#define SYNC	LATDbits.LATD9		//@sync out
#define VIDEO	LATDbits.LATD3		// video out


/*======================================================================
	void video_init(void)
======================================================================*/

void video_init(void)
{
	// sync signal output port
	TRISDbits.TRISD9 = 0; // NTSC sync
	LATDbits.LATD9 = 1;	  // output level High

	// video signal output port
	TRISDbits.TRISD3 = 0; // video out(SDO3)
	LATDbits.LATD3 = 0;	  // output level low

	// set Timer2
	T2CON = 0x0010; // prescale 1:2 40MHz, timer stop
	TMR2 = 0;
	PR2 = H_NTSC-1;
	IEC0SET = _IEC0_T2IE_MASK; // enable interrupt
	IFS0CLR = _IFS0_T2IF_MASK; // flag clear
	IPC2bits.T2IP = 6;         //@priority 6

	// set OC1 interrupt
	IEC0SET = _IEC0_OC1IE_MASK; // enable interrupt
	IFS0CLR = _IFS0_OC1IF_MASK; // flag clear
	IPC1bits.OC1IP = 5;         //@priority 5

	// set OC2 interrupt
	IEC0SET = _IEC0_OC2IE_MASK; // enable interrupt
	IFS0CLR = _IFS0_OC2IF_MASK; // flag clear
	IPC2bits.OC2IP = 5;         //@priority 5

	// set SPI3
	SPI3CON=0x00010834;	// SPI3 disable,32bit,master, not use SDI,
	//SPI3BRG=5;	               // 6.666MHz
	SPI3BRG=4;	                   // 8MHz
	IEC0CLR = _IEC0_SPI3TXIE_MASK; // disable TX interrupt
	IFS0CLR = _IFS0_SPI3TXIF_MASK; // flag clear
	IPC6bits.SPI3IP = 5;           //@priority 5
	SPI3STATbits.SPIROV=0;         // clear ovf bit 
	SPI3CONbits.ON = 1;	           // enable SPI3

	// set video variable
	VState = SV_PREEQ;			   // initialize state
	VCount = V_PREEQ;			   // vertical counter preset
	VPtr = VMem = 0xD000;		   // set video memory begin address
	mnrvs = 0;	                   // selected filename highlite on/off switch

	// go NTSC
	T2CONbits.TON = 1;

}


/*----------------------------------------------------------------------
	Timer2 interrupt
----------------------------------------------------------------------*/

void __attribute__ (( interrupt(ipl6), vector(_TIMER_2_VECTOR) )) _T2Interrupt(void)
{
	SYNC = 0;				// o(RD9)

	// Xe[gɂ蕪
	switch(VState){
	case SV_PREEQ:				// uLO㑤̋
		OC1R = H_SYNC-1;        // pX4.7usec
		OC1CON = 0x8001;		// ^C}2IVbg@L->H
		break;
	case SV_SYNC:			    // Ԓ
		OC1R = H_NTSC-H_SYNC-1; // ؂ꍞ݃pX
		OC1CON = 0x8001;
		break;
	case SV_POSTEQ:				// uLŐ					
		OC1R = H_SYNC-1;        // pX4.7usec
		OC1CON = 0x8001;		// ^C}2IVbg@L->H
		break;
	case SV_LINE: 				// f\
		OC1R = H_SYNC-1;        // pX4.7usec
		OC1CON = 0x8001;		// ^C}2IVbg@L->H
		if(v_gate){
			OC2R = H_SYNC + H_BACK; // fJn܂ł̒x
			OC2CON = 0x8001;		// ^C}2IVbg@L->H
		}
		if(Line & 0x7){
			VPtr = Lptr; // restore begin address of line
		}else{
			Lptr = VPtr; // saved begin address of line
		}
		cgrom=cgROM[(Line++)&0x7]; // set read line
		break;
	}
	//// JE^XV
	VCount--;				// JE^XV
	if(VCount == 0){		// Xe[gIH
		// \t[IH
		if(VState == SV_POSTEQ){
			VPtr = VMem;		// t[̍ŏɍăZbg
			Line = 0;
		}
		VCount = VC[VState];		// CJE^XV
		VState = VS[VState];		// Xe[gXV
		if(VState==SV_PREEQ){
			v_blank = 0;
		}else if(VState==SV_LINE){
			v_blank = 0x80;
		}
	}
	IFS0CLR = _IFS0_T2IF_MASK;		// T2荞݃tONA
}


/*----------------------------------------------------------------------
	OC1 interrupt
----------------------------------------------------------------------*/

void __attribute__ (( interrupt(ipl5), vector(_OUTPUT_COMPARE_1_VECTOR) )) _OC1Interrupt(void)
{
	SYNC = 1;						// pXZbg
	IFS0CLR = _IFS0_OC1IF_MASK;		// OC1荞݃tONA
}


/*----------------------------------------------------------------------
	OC2 interrupt
----------------------------------------------------------------------*/

void __attribute__ (( interrupt(ipl5), vector(_OUTPUT_COMPARE_2_VECTOR) )) _OC2Interrupt(void)
{
	DotCount = 0;
	IEC0SET = _IEC0_SPI3TXIE_MASK;  // SPIM荞݂

	IFS0CLR = _IFS0_OC2IF_MASK;		// OC2荞݃tONA
}


/*----------------------------------------------------------------------
	SPI3 TX interrupt
----------------------------------------------------------------------*/

void __attribute__ (( interrupt(ipl5), vector(_SPI_3_VECTOR) )) _SPI3TXInterrupt(void)
{
	unsigned long d1,d2,d3,d4,data;

	//// rfIMo
	// o͎s(10)
	if(DotCount < 10){		// 40/(4/)
		d1=(cgrom[RAM[VPtr++]])<<24;
		d2=(cgrom[RAM[VPtr++]])<<16;
		d3=(cgrom[RAM[VPtr++]])<<8;
		d4=(cgrom[RAM[VPtr++]]);
		data = d1|d2|d3|d4;

		// file select position ?
		if(mnrvs && Line >= bgnln && Line <= endln &&
   			DotCount >= bgndt && DotCount <= enddt){
			SPI3BUF = scrnRvrs ? data: ~data;
		}else{
			SPI3BUF = scrnRvrs ? ~data: data;
		}

		d1=(cgrom[RAM[VPtr++]])<<24;
		d2=(cgrom[RAM[VPtr++]])<<16;
		d3=(cgrom[RAM[VPtr++]])<<8;
		d4=(cgrom[RAM[VPtr++]]);
		data = d1|d2|d3|d4;

		// file select position ?
		if(mnrvs && Line >= bgnln && Line <= endln &&
   			DotCount >= bgndt && DotCount <= enddt){
			SPI3BUF = scrnRvrs ? data: ~data;
		}else{
			SPI3BUF = scrnRvrs ? ~data: data;
		}

		VPtr &= 0xD7FF; // memory address rolling

		DotCount+=2;					// hbgJE^XV
	}else{
		SPI3BUF = 0x00000000;   // SDO3̏o̓smLowɂ邽߂
		SPI3BUF = 0x00000000;   // zero o͂B
		IEC0CLR = _IEC0_SPI3TXIE_MASK;  // SPIM荞݂֎~
	}

	IFS0CLR = _IFS0_SPI3TXIF_MASK;	// 荞݃tONA
}


/*** end of video.c ****************************************************/



